/* SortedCltn.c -- implementation of sorted collection

	THIS SOFTWARE FITS THE DESCRIPTION IN THE U.S. COPYRIGHT ACT OF A
	"UNITED STATES GOVERNMENT WORK".  IT WAS WRITTEN AS A PART OF THE
	AUTHOR'S OFFICIAL DUTIES AS A GOVERNMENT EMPLOYEE.  THIS MEANS IT
	CANNOT BE COPYRIGHTED.  THIS SOFTWARE IS FREELY AVAILABLE TO THE
	PUBLIC FOR USE WITHOUT A COPYRIGHT NOTICE, AND THERE ARE NO
	RESTRICTIONS ON ITS USE, NOW OR SUBSEQUENTLY.

Author:
	K. E. Gorlen
	Bg. 12A, Rm. 2033
	Computer Systems Laboratory
	Division of Computer Research and Technology
	National Institutes of Health
	Bethesda, Maryland 20892
	Phone: (301) 496-1111
	uucp: uunet!nih-csl!kgorlen
	Internet: kgorlen@alw.nih.gov
	September, 1985

Function:
	
A SortedCltn is a Collection of objects ordered as defined by the
virtual function "compare", which the objects must implement.  The "add"
function locates the position at which to insert the object by
performing a binary search, then invokes the private function
"OrderedCltn::addAtIndex" to insert the object after shifting up all the
objects after it in the array; therefore, a SortedCltn is not efficient
for a large number of objects.

$Log:	SortedCltn.c,v $
 * Revision 3.14  92/12/19  15:53:32  sandy
 * removed extraneous semi-colon
 * 
 * Revision 3.13  92/10/17  16:41:29  sandy
 * corrected return of failed search from findRangeOf
 * 
 * Revision 3.12  92/07/26  14:33:45  sandy
 * include directive in form '#include <nihcl/foo.h>'
 * 
 * Revision 3.11  92/04/30  15:49:29  sandy
 * one initor in DEFINE_CLASS macros
 * 
 * Revision 3.10  91/02/18  21:08:21  kgorlen
 * Release for 3rd printing of 1st edition.
 * 
 * Revision 3.2  90/12/22  08:56:00  kgorlen
 * Fix findIndexOf() to return -1 if SortedCltn empty.
 * 
 * Revision 3.1  90/11/10  21:39:26  kgorlen
 * Eliminate operator=().
 * Change all applications of compare() to apply to contents[]
 * with argument object as argument to compare() to allow ob
 * to not be an association.
 * Eliminate first argument of findIndexOfFirstKey() and
 * findIndexOfLastKey().
 * 
 * Revision 3.0  90/05/20  16:45:35  kgorlen
 * Release for 1st edition.
 * 
*/

#include <nihcl/SortedCltn.h>
#include <nihcl/nihclIO.h>

#define	THIS	SortedCltn
#define	BASE	OrderedCltn
#define BASE_CLASSES BASE::desc()
#define MEMBER_CLASSES
#define VIRTUAL_BASE_CLASSES

DEFINE_CLASS(SortedCltn,1,"$Header: /afs/.alw.nih.gov/system/src/local/nihcl/311/share/lib/RCS/SortedCltn.c,v 3.14 92/12/19 15:53:32 sandy Exp $",NULL)

SortedCltn::SortedCltn(unsigned size) : BASE(size) {}

Object* SortedCltn::add(Object& ob)
{
	if (size()==0) {		// add first object to collection 
		OrderedCltn::add(ob);
		return &ob;
	}

	int i = findIndexOf(ob);

	if (i == -1 || contents[i]->compare(ob) != 0)
		OrderedCltn::addAtIndex(i + 1, ob);
	else
//		An object(s) equal to the argument "ob" already exists.
//		Add another one, after the sequence.
		OrderedCltn::addAtIndex(findIndexOfLastKey(i) + 1, ob);

	return &ob;
}
	
int SortedCltn::findIndexOf(const Object& key) const
{
//	This algorithm is adapted from Horowitz & Sahni,
//	"Fundamentals of Data Structures", 1976, Section 7.1,
//	algorithm BINSRCH.
//	This function will return (on failure to find):
//		(-1) if the key is less than all other keys; 
//		otherwise, returns i, such that i is the greatest index
//		whose key is less than the key value.
//	On successful search the algorithm returns an index to a
//	key which equals the key argument.  However, this is not
//	guaranteed to be the first key of a sequence, when such a
//	sequence exists.

	int l = 0;
	int u = size() - 1;
	int m = 0;
	int c;

	if (u < 0 || contents[0]->compare(key) > 0)
		return -1;

	while (l <= u) {
		m = (l + u) >> 1;
		if ((c = contents[m]->compare(key)) < 0)
			l = m + 1;
		else
		if (c == 0)
			return m;
		else
			u = m - 1;
	}

//	Binary search will leave the final index searched either
//	just greater than the key or just less than, depending 
//	upon the relation of the search key to the existing
//	collection.  Must adjust here, in case it was placed just
//	to the right.

	if (m > 0 && c > 0)
		m--;

	return m; // Key not found.
}

Object* SortedCltn::addAfter(const Object&, Object& /*newob*/)
{
	shouldNotImplement("addAfter"); return 0;
}

Object* SortedCltn::addAllLast(const OrderedCltn&)
{
	shouldNotImplement("addAllLast"); return 0;
}

Object* SortedCltn::addBefore(const Object&, Object& /*newob*/)
{
	shouldNotImplement("addBefore"); return 0;
}

Object* SortedCltn::addLast(Object&)
{
	shouldNotImplement("addLast"); return 0;
}

void SortedCltn::atAllPut(Object&)
{
	shouldNotImplement("atAllPut");
}

int SortedCltn::indexOfSubCollection(const SeqCltn& /*cltn*/, int /*start*/) const
{
	shouldNotImplement("indexOfSubCollection"); return 0;
}

void SortedCltn::replaceFrom(int /*start*/, int /*stop*/, const SeqCltn& /*replacement*/, int /*startAt*/)
{
	shouldNotImplement("replaceFrom");
}

void SortedCltn::sort()
{
	shouldNotImplement("sort");
}
 
SortedCltn Collection::asSortedCltn() const
{
	SortedCltn cltn(MAX(size(),DEFAULT_CAPACITY));
	addContentsTo(cltn);
	return cltn;
}

SortedCltn::SortedCltn(OIOin& strm)
:
#ifdef MI
	Object(strm),
#endif
	BASE(strm)
{
}

SortedCltn::SortedCltn(OIOifd& fd)
:
#ifdef MI
	Object(fd),
#endif
	BASE(fd)
{
}

Range SortedCltn::findRangeOf(const Object& key) const
{
	int i = findIndexOf(key);

	if (i == -1 || contents[i]->compare(key) != 0)
//		Give the caller the place where "key" should
//		be placed, if not found.
//		i.e. either 0 (when i=-1), 
//		     or i+1 where i is largest with value < key.value()
		return Range(i+1, 0);
	int k = findIndexOfFirstKey(i);
	int l = findIndexOfLastKey(i);
	return Range(k, l - k + 1);
}

int SortedCltn::findIndexOfFirstKey(int index) const
{
	const Object& key = *contents[index];
	for (register j = index - 1; j > -1; j--)
		if (contents[j]->compare(key) != 0)
			break;
	return (j + 1);
}

int SortedCltn::findIndexOfLastKey(int index) const
{
	const Object& key = *contents[index];
	register max = size();
	for (register j = index + 1; j < max; j++)
		if (contents[j]->compare(key) != 0)
			break;
	return (j - 1);
}

unsigned SortedCltn::occurrencesOf(const Object& key) const
{
	Range r = findRangeOf(key);
	return unsigned(r.length());
}

Object* SortedCltn::remove(const Object& ob)
{
	Range r = findRangeOf(ob);
	int count = r.length();
	if (count == 0) return nil;
	return removeAtIndex(r.firstIndex());
}
